/****************************************************************************
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *~                             IR Kernel                                    ~
 *~                         Copyright(C) 1993                                ~
 *~                       Software Kinetics Ltd.                             ~
 *~                        All Rights Reserved                               ~
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Copyright Software Kinetics Ltd.
 * 
 * Software Kinetics Ltd. has granted a licence of this software to
 * National Library of Canada.
 *
 * The National Library of Canada grants to the user a royalty-free
 * non-exclusive licence in perpetuity to use, modify, improve,
 * reproduce and sell the software.
 *
 * The software is provided to the user "as is" without warranty of
 * any kind.  Neither Software Kinetics Ltd. nor the National Library
 * of Canada warrants,guarantees or makes any representation regarding
 * the use, or the results of the use of the software in terms of
 * accuracy,reliability, or otherwise.  The entire risk as results and
 * performance of the software is assumed by the user. Neither
 * Software Kinetics Ltd. nor the National Library of Canada shall be
 * liable for any direct, indirect, consequential or incidental
 * damages arising out the use of software.
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ 
 *
 * Program: cspoll.c
 *
 * Purpose: This file contains the function to poll the TCP ports based on
 *          the association list and the function to map the ASN event code
 *          to the IR kernel internal event code.
 *
 * Invocation:
 *
 * Input Arguments:
 *
 * Output/Exit Values: None
 *
 * Functions used:
 *                  cs_PollAssoc()
 *                  ir_map_event()
 *
 * Version: Date:   93/05/12
 *          Name:   Jim Saunders & Benjamin Yen
 *          Reason: First Draft of IR Target Kernel
 *
 * Reference: IR KERNEL - TARGET PRELIMINARY SOFTWARE DESIGN DOCUMENT
 *            IR KERNEL - TARGET SOFTWARE DETAIL DESIGN DOCUMENT
 *            IR KERNEL - INTERFACE REQUIREMENTS SPECIFICATION
 *            OSI - ANSI Z39.50 DOCUMENTATION
 *
 * Notes:
 *
 * $Id: cspoll.c,v 1.1 1993/11/01 17:03:51 irsuprt Exp irsuprt $
 * $Log: cspoll.c,v $
 * Revision 1.1  1993/11/01  17:03:51  irsuprt
 * Initial revision
 *
 *
 *****************************************************************************/

#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/errno.h>
#include <sys/socket.h>
#include "ir.h"
#include "irconst.h"
#include "irerror.h"
#include "irstatab.h"
#include "irassoc.h"

extern int errno;
extern char *sys_errlist[];        /* array of system error messages */

/*****************************************************************************
 *
 * Function: logsockerr
 *
 * Purpose: This function gets the TCP port status and log it in the error
 *          file.
 *
 * Input Arguments:
 *          sd - the TCP port Fd
 *
 * Return values: An error status.
 *
 * Output Arguments:
 *
 * External Variables:
 *
 * Function used:
 *
 * Algorithm:
 *
 * Version: Date:   93/05/12
 *          Name:   Roland Lassalin
 *          Reason: First Draft of IR Target Kernel
 *
 * Reference: IR KERNEL - TARGET PRELIMINARY SOFTWARE DESIGN DOCUMENT
 *            IR KERNEL - TARGET SOFTWARE DETAIL DESIGN DOCUMENT
 *            IR KERNEL - INTERFACE REQUIREMENTS SPECIFICATION
 *            OSI - ANSI Z39.50 DOCUMENTATION
 *
 * Notes:
 *
 ****************************************************************************/
static int 
logsockerr(int sd)
{
    int error = 0;
    int size = sizeof(error);

    if (getsockopt(sd, SOL_SOCKET, SO_ERROR, (char *) &error, &size) == 0)
    {
        if (error)
        {
            sprintf(ERR_STR, "Error bit set for socket %d - socket error = %d\n", sd, error);
            display_error("KERNEL", DISPLAY_USER_ERROR, -1);
        }
    }
    else
    {
        sprintf(ERR_STR, "Error bit set for socket %d - getsockopt error [%s]\n",

#ifdef VMS
                sd, strerror(errno, vaxc$errno));
#else
                sd, sys_errlist[errno]);
#endif

        display_error("KERNEL", DISPLAY_USER_ERROR, -1);
    }

    return error;
}

/*****************************************************************************
 *
 * Function: cs_PollAssoc
 *
 * Purpose: This function polls the following ports on every association:
 *              - the default TCP port for new association,
 *              - the TCP port for the client, and
 *              - the TCP port for the QRM process
 *          for read, write and exception operations.
 *
 * Input Arguments:
 *          serviceSocket - the default TCP port Fd
 *          assoc_list    - pointer to the association list
 *          assocId       - pointer to the current association
 *
 * Return values: An event from the received APDU.
 *
 * Output Arguments:
 *          assocId       - pointer to the current association
 *          PDU_ptr       - the decoded PDU from the received APDU
 *
 * External Variables:
 *
 * Function used:
 *
 * Algorithm:
 *
 * Version: Date:   93/05/12
 *          Name:   Jim Saunders & Benjamin Yen
 *          Reason: First Draft of IR Target Kernel
 *
 * Reference: IR KERNEL - TARGET PRELIMINARY SOFTWARE DESIGN DOCUMENT
 *            IR KERNEL - TARGET SOFTWARE DETAIL DESIGN DOCUMENT
 *            IR KERNEL - INTERFACE REQUIREMENTS SPECIFICATION
 *            OSI - ANSI Z39.50 DOCUMENTATION
 *
 * Notes:
 *
 ****************************************************************************/

int
cs_PollAssoc(serviceSocket, assoc_list, assocId, PDU_ptr)
    int serviceSocket;                /* IR service socket */
    struct queue_struct **assoc_list;
    struct assoc_struct **assocId;
    APDU *PDU_ptr;
{
    fd_set readFds;
    fd_set exceptFds;
    fd_set writeFds;
    int nFound;
    struct timeval timeout;
    int nbytesRead;
    int nbytes;
    int event;
    int apdu_status, return_status;
    FILE *infile;

#ifdef DEBUG_LEVEL1
    fprintf(DEBUG_FILE1, "\nCalling cs_PollAssoc (No Of Assocs = %d)...\n", NO_OF_ASSOC);
#endif

    if ((infile = fopen("irtarget.end", "r")) != NULL)
    {
        ir_close_files();
        exit(1);
    }

    /* Default return event. */
    event = NON_EVENT;

    /* Are there any clients trying to connect */
    FD_ZERO(&readFds);
    FD_ZERO(&exceptFds);
    FD_SET(serviceSocket, &exceptFds);
    FD_SET(serviceSocket, &readFds);

    /* Wait 100 micro seconds */
    timeout.tv_sec = 0;
    timeout.tv_usec = 100;

#ifdef DEBUG_LEVEL1
    fprintf(DEBUG_FILE1, "cs_PollAssoc: polling the service port %d\n", serviceSocket);
#endif

    nFound = select(serviceSocket + 1, &readFds,
                    (fd_set *) 0, &exceptFds, &timeout);

#ifdef DEBUG_LEVEL3
    fprintf(DEBUG_FILE3, "cs_PollAssoc: return value of select on client = %d\n", nFound);
#endif

    if (nFound < 0)
    {
        display_error("KERNEL", ERR_SELECT_TCP_PORT, errno);
        return (FAILED);
    }
    else if (nFound > 0)
    {
        if (FD_ISSET(serviceSocket, &exceptFds))
        {
            /*
             * The peer has aborted. P-Abort. logsockerr() added due to
             * installation of VMS system.
             */
            if (logsockerr(serviceSocket))
            {
                event = AP_ABORT_INDICATION;
                return (event);
            }
        }

        if (FD_ISSET(serviceSocket, &readFds))
        {

#ifdef DEBUG_LEVEL3
            fprintf(DEBUG_FILE3, "cs_PollAssoc: event=A_ASSOCIATE, socket=%d\n", serviceSocket);
#endif

            event = A_ASSOCIATE;
            (*assocId) = ir_assoc_init(assoc_list, serviceSocket);

#ifdef DEBUG_LEVEL3
            fprintf(DEBUG_FILE3, "cs_PollAssoc: after ir_assoc_init: new temp assocID->socket=%d\n", (*assocId)->socket_no[CLIENTSOCK]);
#endif

            return (event);
        }
    }

    /* Just return if there are no established associations */
    if ((*assocId) == NULL)
    {

#ifdef DEBUG_LEVEL1
        fprintf(DEBUG_FILE1, "cs_PollAssoc: assocId is empty, returned\n");
#endif

        return (event);
    }

    /* Check the association's state information. */
    if ((*assocId)->socket_status[CLIENTSOCK] == A_P_ABORTED)
    {

#ifdef DEBUG_LEVEL1
        fprintf(DEBUG_FILE1, "cs_PollAssoc: AP_ABORT_INDICATION\n");
#endif

        event = AP_ABORT_INDICATION;
        return (event);
    }

    /* Poll the client connection */
    if ((*assocId)->socket_status[CLIENTSOCK] == CS_CONNECTED)
    {

#ifdef DEBUG_LEVEL1
        fprintf(DEBUG_FILE1, "cs_PollAssoc: polling the client %d\n", (*assocId)->socket_no[CLIENTSOCK]);
#endif

        FD_ZERO(&readFds);
        FD_ZERO(&exceptFds);
        FD_SET((*assocId)->socket_no[CLIENTSOCK], &exceptFds);
        FD_SET((*assocId)->socket_no[CLIENTSOCK], &writeFds);
        FD_SET((*assocId)->socket_no[CLIENTSOCK], &readFds);

        /* Wait 100 micro seconds */
        timeout.tv_sec = 0;
        timeout.tv_usec = 100;

        nFound = select((*assocId)->socket_no[CLIENTSOCK] + 1, &readFds,
                        (fd_set *) 0, &exceptFds, &timeout);

#ifdef DEBUG_LEVEL3
        fprintf(DEBUG_FILE3, "cs_PollAssoc: return value of select on client = %d\n", nFound);
#endif

        if (nFound < 0)
        {
            display_error("KERNEL", ERR_SELECT_TCP_PORT, errno);
            return (FAILED);
        }
        else if (nFound > 0)
        {
            if (FD_ISSET((*assocId)->socket_no[CLIENTSOCK], &exceptFds))
            {
                /*
                 * The peer has aborted. P-Abort. logsockerr() added due to
                 * installation of VMS system.
                 */
                if (logsockerr((*assocId)->socket_no[CLIENTSOCK]))
                {
                    event = AP_ABORT_INDICATION;
                    return (event);
                }
            }

            /* If something to read then read it. */
            if (FD_ISSET((*assocId)->socket_no[CLIENTSOCK], &readFds))
            {
                apdu_status = cs_receive_apdu(assocId, (*assocId)->socket_no[CLIENTSOCK], &nbytesRead);
                if (apdu_status == APDU_COMPLETE)
                {
                    return_status = cs_Pdata_indication((*assocId)->APDU_ptr, PDU_ptr, (*assocId)->APDUsize);
                    if (return_status == SUCCESS)
                    {
                        event = ir_map_event(PDU_ptr);

#ifdef DEBUG_LEVEL3
                        fprintf(DEBUG_FILE3, "cs_PollAssoc: event = %d\n", event);
#endif        /* return (event); */
                    }
                }
            }
        }
    }

    /* Poll the Target Process connection */
    if ((*assocId)->socket_status[QRMSOCK] == CS_CONNECTED)
    {

#ifdef DEBUG_LEVEL1
        fprintf(DEBUG_FILE1, "cs_PollAssoc: polling the target process %d\n", (*assocId)->socket_no[QRMSOCK]);
#endif

        FD_ZERO(&readFds);
        FD_ZERO(&exceptFds);
        FD_SET((*assocId)->socket_no[QRMSOCK], &exceptFds);
        FD_SET((*assocId)->socket_no[QRMSOCK], &writeFds);
        FD_SET((*assocId)->socket_no[QRMSOCK], &readFds);

        /* Wait 100 micro seconds */
        timeout.tv_sec = 0;
        timeout.tv_usec = 100;

        nFound = select((*assocId)->socket_no[QRMSOCK] + 1, &readFds,
                        (fd_set *) 0, &exceptFds, &timeout);

#ifdef DEBUG_LEVEL3
        fprintf(DEBUG_FILE3, "cs_PollAssoc: return value of select on QRM = %d\n", nFound);
#endif

        if (nFound < 0)
        {
            display_error("KERNEL", ERR_SELECT_TCP_PORT, errno);
            return (FAILED);
        }
        else if (nFound > 0)
        {
            if (FD_ISSET((*assocId)->socket_no[QRMSOCK], &exceptFds))
            {
                /*
                 * The peer has aborted. P-Abort. logsockerr() added due to
                 * installation of VMS system.
                 */
                if (logsockerr((*assocId)->socket_no[QRMSOCK]))
                {
                    event = AP_ABORT_INDICATION;
                    return (event);
                }
            }

            /* If something to read then read it. */
            if (FD_ISSET((*assocId)->socket_no[QRMSOCK], &readFds))
            {

#ifdef DEBUG_LEVEL1
                fprintf(DEBUG_FILE1, "cs_PollAssoc: something to read from QRM \n");
#endif

                apdu_status = cs_receive_apdu(assocId, (*assocId)->socket_no[QRMSOCK], &nbytesRead);

                if (apdu_status == APDU_COMPLETE)
                {
                    return_status = cs_Pdata_indication((*assocId)->APDU_ptr, PDU_ptr, (*assocId)->APDUsize);
                    if (return_status == SUCCESS)
                    {
                        event = ir_map_event(PDU_ptr);

#ifdef DEBUG_LEVEL3
                        fprintf(DEBUG_FILE3, "cs_PollAssoc: event = %d\n", event);
#endif

                        return (event);
                    }
                }
            }

#ifdef DEBUG_LEVEL1
            else
                fprintf(DEBUG_FILE1, "cs_PollAssoc: nothing to read from QRM \n");
#endif
        }
    }

    return (event);

}                                /* end of procedure cs_PollAssoc */

/*****************************************************************************
 *
 * Function: ir_map_event
 *
 * Purpose: This function map the ASN defined events to my IR kernel
 *          defined events. My IR kernel defined events are from the OSI
 *          ANSI Z39.50 state table for target (table 11a and 11b).
 *
 * Input Arguments:
 *          PDU_ptr - pointer to the current association
 *
 * Return values: An event.
 *
 * Output Arguments:
 *
 * External Variables:
 *
 * Function used:
 *
 * Algorithm:
 *
 * Version: Date:   93/05/12
 *          Name:   Benjamin Yen
 *          Reason: First Draft of IR Target Kernel
 *
 * Reference: IR KERNEL - TARGET PRELIMINARY SOFTWARE DESIGN DOCUMENT
 *            IR KERNEL - TARGET SOFTWARE DETAIL DESIGN DOCUMENT
 *            IR KERNEL - INTERFACE REQUIREMENTS SPECIFICATION
 *            OSI - ANSI Z39.50 DOCUMENTATION
 *
 * Notes:
 *
 ****************************************************************************/
int 
ir_map_event(APDU * PDU_ptr)
{
    int new_event;

    switch (PDU_ptr->choiceId)
    {
        case 0:
            new_event = INIT_REQUEST;
            break;
        case 1:
            /* check if init response is okay */
            if (PDU_ptr->a.initResp->result)
                new_event = INIT_RESPONSE_ACCEPT;
            else
                new_event = INIT_RESPONSE_REJECT;
            break;
        case 2:
            new_event = SEARCH_REQUEST;
            break;
        case 3:
            new_event = SEARCH_RESPONSE;
            break;
        case 4:
            new_event = PRESENT_REQUEST;
            break;
        case 5:
            new_event = PRESENT_RESPONSE;
            break;
        case 6:
            new_event = DELETE_REQUEST;
            break;
        case 7:
            new_event = DELETE_RESPONSE;
            break;
        case 8:
            new_event = ACCESS_CONTROL_REQUEST;
            break;
        case 9:
            new_event = ACCESS_CONTROL_RESPONSE;
            break;
        case 10:
            if (PDU_ptr->a.resourceControlReq->responseRequired)
                new_event = RESOURCE_REQUEST_RESP;
            else
                new_event = RESOURCE_REQUEST_NORESP;
            break;
        case 11:
            new_event = RESOURCE_RESPONSE;
            break;
        case 12:
            new_event = TRIGGER_RESOURCE_REQUEST;
            break;
        case 13:
            new_event = RESOURCE_REPORT_REQUEST;
            break;
        case 14:
            new_event = RESOURCE_REPORT_RESPONSE;
            break;
    }

    return (new_event);
}
