/****************************************************************************
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *~                             IR Kernel                                    ~
 *~                         Copyright(C) 1993                                ~
 *~                       Software Kinetics Ltd.                             ~
 *~                        All Rights Reserved                               ~
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Copyright Software Kinetics Ltd.
 * 
 * Software Kinetics Ltd. has granted a licence of this software to
 * National Library of Canada.
 *
 * The National Library of Canada grants to the user a royalty-free
 * non-exclusive licence in perpetuity to use, modify, improve,
 * reproduce and sell the software.
 *
 * The software is provided to the user "as is" without warranty of
 * any kind.  Neither Software Kinetics Ltd. nor the National Library
 * of Canada warrants,guarantees or makes any representation regarding
 * the use, or the results of the use of the software in terms of
 * accuracy,reliability, or otherwise.  The entire risk as results and
 * performance of the software is assumed by the user. Neither
 * Software Kinetics Ltd. nor the National Library of Canada shall be
 * liable for any direct, indirect, consequential or incidental
 * damages arising out the use of software.
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ 
 *
 * Program: csrecv.c
 *
 * Purpose: This file contains the function to receive APDU with a selected 
 *          socketFd.
 *
 * Invocation:
 *
 * Input Arguments:
 *
 * Output/Exit Values: None
 *
 * Functions used: cs_receive_apdu()
 *
 * Version: Date:   93/05/12
 *          Name:   Jim Saunders & Benjamin Yen
 *          Reason: First Draft of IR Target Kernel
 *
 * Reference: IR KERNEL - TARGET PRELIMINARY SOFTWARE DESIGN DOCUMENT
 *            IR KERNEL - TARGET SOFTWARE DETAIL DESIGN DOCUMENT
 *            IR KERNEL - INTERFACE REQUIREMENTS SPECIFICATION
 *            OSI - ANSI Z39.50 DOCUMENTATION
 *
 * Notes:
 *
 * $Id: csrecv.c,v 1.2 1993/12/13 19:17:37 irsuprt Exp irsuprt $
 * $Log: csrecv.c,v $
 * Revision 1.2  1993/12/13  19:17:37  irsuprt
 * V1.2: added #include <error.h> for AIX_SYSTEM.
 *
 * Revision 1.1  1993/11/01  17:03:51  irsuprt
 * Initial revision
 *
 *
 *****************************************************************************/
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#ifdef AIX_SYSTEM
	#include <error.h>
#endif 
#include "ir.h"
#include "irerror.h"
#include "irassoc.h"
#include "irstatab.h"
#include "irconst.h"
#include "asn_incl.h" 

/* #define TEMP_DEBUG */

#ifdef TEMP_DEBUG
	int gi=1;
#endif

/*****************************************************************************
 *
 * Function: cs_receive_apdu
 *
 * Purpose: To receive an APDU from a TCP data stream. It detects if there is
 *          a second APDU within the received APDU. If it does, it will make
 *          a copy and stored in the second APDU pointer Id APDU_ptr2 and
 *          the second APDU size Id APDUsize2. The function ir_control() will
 *          detect the second APDU and then passed to ACSE() or ir_ASE().
 *
 * Input Arguments: assocId    - pointer to the current association
 *                  socket     - socket file Id 
 *
 * External Variables:
 *
 * Return values: APDU_COMPLETE   - APDU received in full or successfully
 *                APDU_INCOMPLETE - APDU received not in full or failed
 *
 * Output Arguments: nbytesRead - number of bytes received 
 *
 * External Variables:
 *
 * Function used:
 *                  recv()
 *                  SBufInstallData()
 *                  SBufResetInReadMode()
 *                  setjmp()
 *                  BDecTag()
 *                  BDecLen()
 *                  display_error()
 *
 * Algorithm:
 *
 * Version: Date:   93/05/12
 *          Name:   Jim Saunders & Benjamin Yen
 *          Reason: First Draft of IR Target Kernel
 *
 * References:
 *
 * Notes:
 *
 ****************************************************************************/

int cs_receive_apdu( assocId, socket, nbytesRead )
struct assoc_struct **assocId;
int socket;
int *nbytesRead;
{
    int val;
    AsnLen bsize, bsizea;
    char **buf;
    int remainingBytes;
    ENV_TYPE env;
    SBuf outputBuf;
    AsnTag tag;
    APDU apduptr;
	FILE *f;
	char str[15];

#ifdef DEBUG_LEVEL1
    fprintf(DEBUG_FILE1, "Calling cs_receive_apdu\n");
    fprintf(DEBUG_FILE1, "cs_receive_apdu: next_pos=%ld, APDUsize=%ld\n", (*assocId)->next_start_pos, (*assocId)->APDUsize);
#endif

#ifdef DEBUG_LEVEL3
    fprintf(DEBUG_FILE3, "cs_receive_apdu: next_pos=%ld, APDUsize=%ld\n", (*assocId)->next_start_pos, (*assocId)->APDUsize);
#endif

    /* Is this the start of a new APDU ? */
    if ((*assocId)->next_start_pos >= (*assocId)->APDUsize) 
    {

        *nbytesRead = recv( socket, (*assocId)->APDU_ptr, MAXMESSAGESIZE, 0);

#ifdef DEBUG_LEVEL3
    fprintf(DEBUG_FILE3, "cs_receive_apdu: nbytesRead=%d\n",*nbytesRead);
#endif

#ifdef TEMP_DEBUG
	sprintf(str, "rec%-d.pdu", gi++);
	f=fopen(str, "wb");
	fwrite((*assocId)->APDU_ptr, *nbytesRead, 1, f);
	fclose(f);
#endif

        if (*nbytesRead <= 0)
            return(APDU_INCOMPLETE);

        memcpy((*assocId)->working_data_buf, (*assocId)->APDU_ptr, MAXMESSAGESIZE*2);

        /* Find out the actual size of the APDU */
        SBufInstallData (&outputBuf, (*assocId)->working_data_buf, *nbytesRead);     
        SBufResetInReadMode(&outputBuf); 
   
        bsizea = (AsnLen) 0; 
        if ((val = setjmp( env ) ) == 0 )
        {
            tag = BDecTag( &outputBuf, &bsizea, env );
            bsize = BDecLen(&outputBuf, &bsizea, env); 
            (*assocId)->APDUsize = bsize + bsizea;
            (*assocId)->next_start_pos = *nbytesRead; 

            /* if there is a second APDU */
            if ((*nbytesRead - (*assocId)->APDUsize) > 0)
            {
                (*assocId)->APDUsize2 = *nbytesRead - (*assocId)->APDUsize;
#ifdef DEBUG_LEVEL3
    fprintf(DEBUG_FILE3, "cs_receive_apdu: *assocId APDUsize2 %ld\n",(*assocId)->APDUsize2);
#endif
                memcpy((*assocId)->APDU_ptr2, (*assocId)->APDU_ptr+(*assocId)->APDUsize, (*assocId)->APDUsize2);
            }

#ifdef DEBUG_LEVEL3
    fprintf(DEBUG_FILE3, "cs_receive_apdu: *assocId APDUsize %ld\n",(*assocId)->APDUsize);
#endif

        } 
        else
        {
            display_error("KERNEL", ERR_FIND_APDU_LENGTH, errno);
            return(APDU_INCOMPLETE);
        }
    }
    else /* appending to the existing APDU */
    {
        remainingBytes = (*assocId)->APDUsize - (*assocId)->next_start_pos;    
        if (remainingBytes > MAXMESSAGESIZE)
            remainingBytes = MAXMESSAGESIZE;
        *nbytesRead = recv( socket, ((*assocId)->APDU_ptr+(*assocId)->next_start_pos), remainingBytes, 0);

        (*assocId)->next_start_pos += (long) *nbytesRead;

#ifdef DEBUG_LEVEL3
    fprintf(DEBUG_FILE3, "cs_receive_apdu: appending nbytesRead=%d\n",*nbytesRead);
#endif

    }

    if ((*assocId)->next_start_pos >= (*assocId)->APDUsize)
    {
        return(APDU_COMPLETE); 
    }
    else
    {
        return(APDU_INCOMPLETE);
    }
}  /* end of procedure cs_receive_apdu  */
