/****************************************************************************
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *~                             IR Kernel                                    ~
 *~                         Copyright(C) 1993                                ~
 *~                       Software Kinetics Ltd.                             ~
 *~                        All Rights Reserved                               ~
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Copyright Software Kinetics Ltd.
 * 
 * Software Kinetics Ltd. has granted a licence of this software to
 * National Library of Canada.
 *
 * The National Library of Canada grants to the user a royalty-free
 * non-exclusive licence in perpetuity to use, modify, improve,
 * reproduce and sell the software.
 *
 * The software is provided to the user "as is" without warranty of
 * any kind.  Neither Software Kinetics Ltd. nor the National Library
 * of Canada warrants,guarantees or makes any representation regarding
 * the use, or the results of the use of the software in terms of
 * accuracy,reliability, or otherwise.  The entire risk as results and
 * performance of the software is assumed by the user. Neither
 * Software Kinetics Ltd. nor the National Library of Canada shall be
 * liable for any direct, indirect, consequential or incidental
 * damages arising out the use of software.
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ 
 *
 * Program: irrwatch.c
 *
 * Purpose: This file contains the functions to monitor the timeout for the
 *          response event and idle association.
 *
 * Invocation:
 *
 * Input Arguments:
 *
 * Output/Exit Values: None
 *
 * Functions used: ir_response_watchdog()
 *                 ir_alarm_expires()
 *                 ir_idle_too_long()
 *
 * Version: Date:   93/05/12
 *          Name:   Benjamin Yen
 *          Reason: First Draft of IR Target Kernel
 *
 * Reference: IR KERNEL - TARGET PRELIMINARY SOFTWARE DESIGN DOCUMENT
 *            IR KERNEL - TARGET SOFTWARE DETAIL DESIGN DOCUMENT
 *            IR KERNEL - INTERFACE REQUIREMENTS SPECIFICATION
 *            OSI - ANSI Z39.50 DOCUMENTATION
 *
 * Notes:
 *
 * $Id: irrwatch.c,v 1.1 1993/11/01 17:03:51 irsuprt Exp irsuprt $
 * $Log: irrwatch.c,v $
 * Revision 1.1  1993/11/01  17:03:51  irsuprt
 * Initial revision
 *
 *
 *****************************************************************************/

#include <stdio.h>
#include <time.h>
#include <sys/errno.h>
#include "ir.h"
#include "irstatab.h"
#include "irconst.h"
#include "irstksta.h"
#include "irstkser.h"
#include "irassoc.h"
#include "irerror.h"

extern int errno;

/*****************************************************************************
 *
 * Function: ir_response_watchdog
 *
 * Purpose: This function activates or deactivates the response timer by
 *          pushing a new node with the mature time to the service stack or
 *          poping the response node from the stack. The function
 *          ir_alarm_expires() checks if the timer has expired.
 *
 * Input Arguments:
 *          assocId         - pointer to the current association
 *          timeout_in_secs - timeout for a particular response in seconds
 *          is_activate     - ACTIVATE or DEACTIVATE the timer
 *
 * Return values:   SUCCESS - operation was completed successfully
 *                  FAILED  - operation failed
 *
 * Output Arguments:
 *
 * External Variables:
 *
 * Function used:
 *                  ir_push_service()
 *                  display_error()
 *                  ir_add_time_service()
 *                  ir_pop_service()
 *
 * Algorithm:
 *
 * Version: Date:   93/05/12
 *          Name:   Benjamin Yen
 *          Reason: First Draft of IR Target Kernel
 *
 * Reference:
 *
 * Notes:
 *
 ****************************************************************************/
int
ir_response_watchdog(assocId, timeout_in_secs, is_activate)
    struct assoc_struct **assocId;
    int timeout_in_secs, is_activate;
{
    time_t mature_time;
    struct tm *today_time;

    if (is_activate == ACTIVATE)
    {
        mature_time = time(NULL);
        today_time = localtime(&mature_time);

#ifdef DEBUG_LEVEL3
        fprintf(DEBUG_FILE3, "Time Now: %d:%d:%d\n", today_time->tm_hour, today_time->tm_min, today_time->tm_sec);
#endif

        mature_time += timeout_in_secs;
        today_time = localtime(&mature_time);

#ifdef DEBUG_LEVEL3
        fprintf(DEBUG_FILE3, "Mature Time: %d:%d:%d=%ld\n", today_time->tm_hour, today_time->tm_min,
                today_time->tm_sec, mature_time);
#endif

        if (ir_push_service(&(*assocId)->stack_service_head, mature_time) == FAILED)
        {
            display_error("KERNEL", ERR_ALLOCATE_MEMORY, errno);
            return (FAILED);
        }
        else
        {
            ir_add_time_service(&(*assocId)->stack_service_head, (long) timeout_in_secs);
        }
    }
    else
    {
        /* deactivate the timing for a particular request */
        ir_pop_service(&(*assocId)->stack_service_head);
    }

    return (SUCCESS);
}

/*****************************************************************************
 *
 * Function: ir_alarm_expires
 *
 * Purpose: This function checks the response mature time on the service
 *          stack against the current time to see if it's expired.
 *
 * Input Arguments:
 *          assocId         - pointer to the current association
 *
 * Return values:   TRUE    - the response timer is expired.
 *                  FALSE   - not expired, wait a little longer.
 *
 * Output Arguments:
 *
 * External Variables:
 *
 * Function used:
 *                  ir_top_service()
 *
 * Algorithm:
 *
 * Version: Date:   93/05/12
 *          Name:   Benjamin Yen
 *          Reason: First Draft of IR Target Kernel
 *
 * Reference:
 *
 * Notes:
 *
 ****************************************************************************/
ir_alarm_expires(assocId)
    struct assoc_struct **assocId;
{
    time_t time_now, mature_time;

    time_now = time(NULL);
    mature_time = ir_top_service(&(*assocId)->stack_service_head);

    if (mature_time == 0)
        return (FALSE);

    if (mature_time < time_now)
        return (TRUE);
    else
        return (FALSE);
}

/*****************************************************************************
 *
 * Function: ir_idle_too_long
 *
 * Purpose: This function checks the last event time Id last_event_time against
 *          the current time to see if its idle time exceeds
 *          IDLE_TIMEOUT_IN_SECONDS.
 *
 * Input Arguments:
 *          assocId         - pointer to the current association
 *
 * Return values:   TRUE    - idle too long.
 *                  FALSE   - can wait a little longer.
 *
 * Output Arguments:
 *
 * External Variables:
 *
 * Function used:
 *                  ir_top_service()
 *
 * Algorithm:
 *
 * Version: Date:   93/05/12
 *          Name:   Benjamin Yen
 *          Reason: First Draft of IR Target Kernel
 *
 * Reference:
 *
 * Notes:
 *
 ****************************************************************************/
ir_idle_too_long(assocId)
    struct assoc_struct **assocId;
{
    time_t time_now, idle_time, idle_timeout;

    time_now = time(NULL);

    /* if this association is waiting for a response, reset last_event_time */
    if ((*assocId)->stack_service_head != NULL)
    {
        (*assocId)->last_event_time = time_now;
        return (FALSE);
    }

    idle_time = time_now - (*assocId)->last_event_time;
    idle_timeout = IDLE_TIMEOUT_IN_SECONDS;

#ifdef DEBUG_LEVEL3
    fprintf(DEBUG_FILE3, "idle_time=%ld > idle_timeout=%ld\n", idle_time, IDLE_TIMEOUT_IN_SECONDS);
#endif

    if (idle_time > IDLE_TIMEOUT_IN_SECONDS)
        return (TRUE);
    else
        return (FALSE);
}
