/****************************************************************************
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *~                             IR Kernel                                    ~
 *~                         Copyright(C) 1993                                ~
 *~                       Software Kinetics Ltd.                             ~
 *~                        All Rights Reserved                               ~
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Copyright Software Kinetics Ltd.
 * 
 * Software Kinetics Ltd. has granted a licence of this software to
 * National Library of Canada.
 *
 * The National Library of Canada grants to the user a royalty-free
 * non-exclusive licence in perpetuity to use, modify, improve,
 * reproduce and sell the software.
 *
 * The software is provided to the user "as is" without warranty of
 * any kind.  Neither Software Kinetics Ltd. nor the National Library
 * of Canada warrants,guarantees or makes any representation regarding
 * the use, or the results of the use of the software in terms of
 * accuracy,reliability, or otherwise.  The entire risk as results and
 * performance of the software is assumed by the user. Neither
 * Software Kinetics Ltd. nor the National Library of Canada shall be
 * liable for any direct, indirect, consequential or incidental
 * damages arising out the use of software.
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ 
 *
 * Program: irtarget.c
 *
 * Purpose: This is the main program which implements the IR Target
 *          Kernel OSI ANSI Z39.50 Protocol. This IR kernel consists of
 *          control function (communication dispatcher),
 *          ACSE(Association Control Service Elements), and IR ASE(Application
 *          Service Element).
 *
 * Invocation: irtarget [port=9999] [host=hostname]
 *
 * Input Arguments:
 *                target port number - target process TCP port number
 *              target host name - target process host name
 *
 * Output/Exit Values: None
 *
 * Functions used:
 *                 ir_state_transition()
 *                 ir_ASE()
 *                 ira_comm_event()
 *                 ACSE()
 *                 ir_open_files()
 *                 ir_close_files()
 *                 ir_get_parameters()
 *                 ir_control()
 *
 * Version: Date:   93/05/12
 *          Name:   Benjamin Yen
 *          Reason: First Draft of IR Target Kernel
 *
 * Reference: IR KERNEL - TARGET PRELIMINARY SOFTWARE DESIGN DOCUMENT
 *            IR KERNEL - TARGET SOFTWARE DETAIL DESIGN DOCUMENT
 *            IR KERNEL - INTERFACE REQUIREMENTS SPECIFICATION
 *            OSI - ANSI Z39.50 DOCUMENTATION
 *
 * Notes:
 *
 * $Id: irtarget.c,v 1.6 1993/12/14 14:19:04 irsuprt Exp irsuprt $
 * $Log: irtarget.c,v $
 * Revision 1.6  1993/12/14  14:19:04  irsuprt
 * V1.2: changed version to 1.2.
 *
 * Revision 1.5  1993/12/13  20:33:06  irsuprt
 * V1.2: Undo 1.3 & 1.4.
 *
 * Revision 1.2  1993/11/04  20:12:00  irsuprt
 * added the display of the source code version
 *
 * Revision 1.1  1993/11/01  17:03:51  irsuprt
 * Initial revision
 *
 *
 *****************************************************************************/

#include <stdio.h>
#include <time.h>
#include <string.h>
#include <stdlib.h>

#include "ir.h"
#include "irstatab.h"
#include "irassoc.h"
#include "irconst.h"
#include "iraccctr.h"
#include "acse.h"
#include "irdelete.h"
#include "irinit.h"
#include "iraliset.h"
#include "ircomevt.h"
#include "irerror.h"
#include "irpresnt.h"
#include "irrsctrl.h"
#include "irrsctrp.h"
#include "irsearch.h"
#include "irterm.h"
#include "irtgrsct.h"
#include "irstksta.h"
#include "irstkser.h"
#include "irrwatch.h"
#include "asn_incl.h"

#define MAXNAMELENGTH   35
#define Z3950_PORT      2007

/********** State Table IDs Definition ************/

/* IR Target State Table Definitions */

/* state table definition */
/* state = 99 means it should get a state from a stack */
int stateTable[EVENTS][STATES] =
{
     /* State                  0  1  2  3  4  5  6  7  8  9 10 11 */
     /* Non-Entry         */ {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1},
     /* 01 InitPDU        */ {-1, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
     /* 02 InitRespAcep   */ {-1, 0, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0},
     /* 03 InitRespRej    */ {-1, 0,11, 0, 0, 0, 0, 0, 0, 0, 0, 0},
     /* 04 SrchPDU        */ {-1, 0, 0, 4, 0, 0, 0, 0, 0, 0, 0, 0},
     /* 05 SrchResp       */ {-1, 0, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0},
     /* 06 PrsntPDU       */ {-1, 0, 0, 5, 0, 0, 0, 0, 0, 0, 0, 0},
     /* 07 PrsntResp      */ {-1, 0, 0, 0, 0, 3, 0, 0, 0, 0, 0, 0},
     /* 08 DltePDU        */ {-1, 0, 0, 6, 0, 0, 0, 0, 0, 0, 0, 0},
     /* 09 DlteResp       */ {-1, 0, 0, 0, 0, 0, 3, 0, 0, 0, 0, 0},
     /* 10 RsrpPDU        */ {-1, 0, 0, 7, 0, 0, 0, 0, 0, 0, 0, 0},
     /* 11 Rsrpresp       */ {-1, 0, 0, 0, 0, 0, 0, 3, 0, 0, 0, 0},
     /* 12 TrigrcPDU      */ {-1, 0, 2, 3, 4, 5, 6, 0, 8, 9, 0,11},
     /* 13 Rscreq(Resp)   */ {-1, 0, 8, 0, 8, 8, 8, 0, 0, 0, 0, 0},
     /* 14 Rscreq(NoResp) */ {-1, 0, 2, 0, 4, 5, 6, 0, 0, 0, 0, 0},
     /* 15 RscrespPDU     */ {-1, 0, 0, 0, 0, 0, 0, 0,99, 0, 0, 0},
     /* 16 Accreq         */ {-1, 0, 9, 0, 9, 9, 9, 9, 0, 0, 0, 0},
     /* 17 AccrespPDU     */ {-1, 0, 0, 0, 0, 0, 0, 0, 0,99, 0, 0},
     /* 18 ArelInd        */ {-1, 0, 0,10, 0, 0, 0, 0, 0, 0, 0,10},
     /* 19 IrelResp       */ {-1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0},
     /* 20 AabInd         */ {-1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1},
     /* 21 APabInd        */ {-1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1},
     /* 22 IabReq         */ {-1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1},
};

/********** End Of State Table Definition **********/

/********** G L O B A L S  D E F I N E D **********/

long MAXMESSAGESIZE;
long MAX_ERROR_RECORDS;

char SOURCE_CODE_VERSION[] = "2.00";

int ACCESS_CONTROL_REQUIRED, RESOURCE_CONTROL_REQUIRED;

FILE *DEBUG_FILE1, *DEBUG_FILE2, *DEBUG_FILE3, *DEBUG_FILE4;
char DEBUG_FILE_NAME1[35], DEBUG_FILE_NAME2[35], DEBUG_FILE_NAME3[35];
char ERROR_FILE_NAME[35];

int INIT_TIMEOUT_IN_SECONDS;
int SEARCH_TIMEOUT_IN_SECONDS;
int PRESENT_TIMEOUT_IN_SECONDS;
int DELETE_TIMEOUT_IN_SECONDS;
int TRIGGER_TIMEOUT_IN_SECONDS;
int ACCESS_TIMEOUT_IN_SECONDS;
int RESOURCE_TIMEOUT_IN_SECONDS;
int IDLE_TIMEOUT_IN_SECONDS;

int NO_OF_ASSOC;

char *PDU;                        /* dummy for testing */

char TEMP_HOST_NAME[35];

/********** E N D   O F   G L O B A L S  *********/

/**********  M A C R O S   D E F I N E D *********/

#define confirmed_ACSE(event)  ((event >= ACSE_EVENT) ? TRUE : FALSE)
#define confirmed_IR(event)  ((event <  ACSE_EVENT) ? TRUE : FALSE)

/*********** E N D  O F  M A C R O S  **********/


/*****************************************************************************
 *
 * Function: ir_state_transition
 *
 * Purpose: This function returns the next state based on the input parameters
 *          present_state and event. If they fail to satisfy the IR kernel
 *          state table based on ANSI Z39.50 protocol, the third parameter
 *          protocol_error will indicate a protocol error.
 *
 * Input Arguments: present_state - present state.
 *                  event - incoming event from origin or target process.
 *
 * Return values: The next state.
 *
 * Output Arguments: protocol_error - a returned parameter to indicate if a
 *                                    protocol error occurs.
 * External Variables:
 *
 * Function used:
 *
 * Algorithm: Based on ISO ANSI Z39.50 IR kernel state machine.
 *
 * Version: Date:   93/05/12
 *          Name:   Benjamin Yen
 *          Reason: First Draft of IR Target Kernel
 *
 * Reference: IR KERNEL - TARGET PRELIMINARY SOFTWARE DESIGN DOCUMENT
 *            IR KERNEL - TARGET SOFTWARE DETAIL DESIGN DOCUMENT
 *            IR KERNEL - INTERFACE REQUIREMENTS SPECIFICATION
 *            OSI - ANSI Z39.50 DOCUMENTATION
 *
 * Notes:
 *
 ****************************************************************************/

int
ir_state_transition(present_state, event, assocId, protocol_error)
    int present_state;
    int event;
    struct assoc_struct **assocId;
    int *protocol_error;
{

    int nextState, currentState;

#ifdef DEBUG_LEVEL1
    fprintf(DEBUG_FILE1, "Calling ir_state_transition ...\n");
#endif

    /* range check, then nextState get the next state from the state table */
    if (((present_state > 0) && (present_state < STATES)) &&
            ((event > 0) && (event < EVENTS))
        )
    {
        nextState = stateTable[event][present_state];
    }

    if (nextState == PROTOCOL_FAILED)
    {
        *protocol_error = TRUE;
        currentState = present_state;
    }
    else
    {
        *protocol_error = FALSE;

        /*
         * save the current state to a stack if the event is resource control
         * or access control and the current states is init, search,
         * present or delete receive.
         */
        if (((present_state == INIT_RECVD || present_state == SEARCH_RECVD ||
            present_state == PRESENT_RECVD || present_state == DELETE_RECVD) &&
            (event == RESOURCE_REQUEST_RESP || event == ACCESS_CONTROL_REQUEST))
            ||
            (present_state == RESOURCE_REPORT_RECVD && event == ACCESS_CONTROL_REQUEST)
           )
            ir_push_state(&(*assocId)->stack_state_head, present_state);

        /*
         * if the next state is a pop state, currentState gets the next state
         * from the stack. Otherwise, the currentState gets from the state
         * table.
         */
        if (nextState == POP_STATE)
            currentState = ir_pop_state(&(*assocId)->stack_state_head);
        else
            currentState = nextState;
    }

    return (currentState);

}                                /* end of procedure ir_state_transition   */

/*****************************************************************************
 *
 * Function: ir_ASE
 *
 * Purpose: This function checks the incoming event against the IR protocol
 *          state machine. If it's valid then it calls the appropriate
 *          function.
 *
 * Input Arguments:      assoc_list      - pointer to the association list
 *                       assocId         - pointer to the current association
 *                       PDU_ptr         - pointer to the current APDU
 *                       event           - current event
 *                       tp_hostname     - target process hostname
 *
 * Return values:        SUCCESS - successfully executed.
 *                       FAILED  - internal target process error has occurred
 *                       PROTOCOL_FAILED  - the event is invalid based on the
 *                                          state machine.
 *
 * Output Arguments:
 *
 *
 * Function called:      ir_state_transition()
 *                       ir_init_request()
 *                       ir_init_response()
 *                       ir_reject()
 *                       ir_search_request()
 *                       ir_search_response()
 *                       ir_present_request()
 *                       ir_present_response()
 *                       ir_delete_request()
 *                       ir_delete_response()
 *                       ir_resource_report_request()
 *                       ir_resource_report_response()
 *                       ir_trigger_resource_control_request()
 *                       ir_resource_control_request()
 *                       ir_resource_control_response()
 *                       ir_access_control_request()
 *                       ir_access_control_response()
 *                       ir_release_request()
 *                       ir_release_response()
 *                       ira_abort_request()
 *                       ir_response_watchdog()
 *
 * Version: Date:   93/03/14
 *          Name:   Benjamin Yen
 *          Reason: First Draft of IR Target Kernel
 *          J.S.    93/03/21 Revised to support multi-threading
 *
 * Reference: IR KERNEL - TARGET PRELIMINARY SOFTWARE DESIGN DOCUMENT
 *            IR KERNEL - TARGET SOFTWARE DETAIL DESIGN DOCUMENT
 *
 * Notes:
 *
 ****************************************************************************/

int
ir_ASE(assoc_list, assocId, PDU_ptr, event, tp_hostname)
    struct queue_struct **assoc_list;
    struct assoc_struct **assocId;
    APDU *PDU_ptr;
    int event;
    char *tp_hostname;
{

    int protocol_error, result;

#ifdef DEBUG_LEVEL1
    fprintf(DEBUG_FILE1, "Calling ir_ASE ...\n");
#endif

#ifdef DEBUG_LEVEL3
    fprintf(DEBUG_FILE3, "ir_ASE:before transition: socket=%d, event=%d, present state=%d, protocol_error=%d\n", (*assocId)->socket_no[CLIENTSOCK], event, (*assocId)->present_state, protocol_error);
#endif

    /* What is the next state ? */
    (*assocId)->present_state = ir_state_transition((*assocId)->present_state, event, assocId, &protocol_error);
    (*assocId)->last_event = event;

#ifdef DEBUG_LEVEL3
    fprintf(DEBUG_FILE3, "ir_ASE:after transition: socket=%d, event=%d, present state=%d, protocol_error=%d\n", (*assocId)->socket_no[CLIENTSOCK], event, (*assocId)->present_state, protocol_error);
#endif

    /* Process the IR event */
    if (!protocol_error)
        switch (event)
        {

            case INIT_REQUEST:
                result = ir_init_request(assocId, PDU_ptr, tp_hostname);
                if (result == SUCCESS)
                {
                    result = ir_response_watchdog(assocId, INIT_TIMEOUT_IN_SECONDS, ACTIVATE);
                }
                break;

            case INIT_RESPONSE_ACCEPT:
                result = ir_init_response(assocId, PDU_ptr);
                if (result == SUCCESS)
                {
                    result = ir_response_watchdog(assocId, NOT_USED, DEACTIVATE);
                }
                break;

            case INIT_RESPONSE_REJECT:
                result = ir_init_response(assocId, PDU_ptr);
                result = ir_reject(assocId, PDU_ptr);
                if (result == SUCCESS)
                {
                    result = ir_response_watchdog(assocId, NOT_USED, DEACTIVATE);
                }
                result = ira_abort_request(assocId);
                break;

            case SEARCH_REQUEST:
                result = ir_search_request(assocId, PDU_ptr);
                if (result == SUCCESS)
                {
                    result = ir_response_watchdog(assocId, SEARCH_TIMEOUT_IN_SECONDS, ACTIVATE);
                }
                break;

            case SEARCH_RESPONSE:
                result = ir_search_response(assocId, PDU_ptr);
                if (result == SUCCESS)
                {
                    result = ir_response_watchdog(assocId, NOT_USED, DEACTIVATE);
                }
                break;

            case PRESENT_REQUEST:
                result = ir_present_request(assocId, PDU_ptr);
                if (result == SUCCESS)
                {
                    result = ir_response_watchdog(assocId, PRESENT_TIMEOUT_IN_SECONDS, ACTIVATE);
                }
                break;

            case PRESENT_RESPONSE:
                result = ir_present_response(assocId, PDU_ptr);
                if (result == SUCCESS)
                {
                    result = ir_response_watchdog(assocId, NOT_USED, DEACTIVATE);
                }
                break;

            case DELETE_REQUEST:
                result = ir_delete_request(assocId, PDU_ptr);
                if (result == SUCCESS)
                {
                    result = ir_response_watchdog(assocId, DELETE_TIMEOUT_IN_SECONDS, ACTIVATE);
                }
                break;

            case DELETE_RESPONSE:
                result = ir_delete_response(assocId, PDU_ptr);
                if (result == SUCCESS)
                {
                    result = ir_response_watchdog(assocId, NOT_USED, DEACTIVATE);
                }
                break;

            case RESOURCE_REPORT_REQUEST:
                result = ir_resource_report_request(assocId, PDU_ptr);
                if (result == SUCCESS)
                {
                    result = ir_response_watchdog(assocId, RESOURCE_TIMEOUT_IN_SECONDS, ACTIVATE);
                }
                break;

            case RESOURCE_REPORT_RESPONSE:
                result = ir_resource_report_response(assocId, PDU_ptr);
                if (result == SUCCESS)
                {
                    result = ir_response_watchdog(assocId, NOT_USED, DEACTIVATE);
                }
                break;

            case TRIGGER_RESOURCE_REQUEST:
                result = ir_trigger_resource_control_request(assocId, PDU_ptr);
                if (result == SUCCESS)
                {
                    result = ir_response_watchdog(assocId, TRIGGER_TIMEOUT_IN_SECONDS, ACTIVATE);
                }
                break;

            case RESOURCE_REQUEST_RESP:
                result = ir_resource_control_request(assocId, PDU_ptr);
                if (result == SUCCESS)
                {
                    result = ir_response_watchdog(assocId, RESOURCE_TIMEOUT_IN_SECONDS, ACTIVATE);
                }
                break;

            case RESOURCE_REQUEST_NORESP:
                result = ir_resource_control_request(assocId, PDU_ptr);
                break;

            case RESOURCE_RESPONSE:
                result = ir_resource_control_response(assocId, PDU_ptr);
                if (result == SUCCESS)
                {
                    result = ir_response_watchdog(assocId, NOT_USED, DEACTIVATE);
                }
                break;

            case ACCESS_CONTROL_REQUEST:
                result = ir_access_control_request(assocId, PDU_ptr);
                if (result == SUCCESS)
                {
                    result = ir_response_watchdog(assocId, ACCESS_TIMEOUT_IN_SECONDS, ACTIVATE);
                }
                break;

            case ACCESS_CONTROL_RESPONSE:
                result = ir_access_control_response(assocId, PDU_ptr);
                if (result == SUCCESS)
                {
                    result = ir_response_watchdog(assocId, NOT_USED, DEACTIVATE);
                }
                break;

            case A_RELEASE_INDICATION:
                result = ir_release_request(assocId);

                break;

            case IR_RELEASE_RESPONSE:        /* **** is a response necessary ? *** */
                result = ir_release_response(assocId);
                break;

            case A_ABORT_INDICATION:
            case AP_ABORT_INDICATION:
            case IR_ABORT_REQUEST:
                result = ira_abort_request(assocId);
                break;

            default:
                result = ira_abort_request(assocId);
                protocol_error = TRUE;
                break;
        }

	ResetNibbleMem(MAXMESSAGESIZE, MAXMESSAGESIZE);

    if (protocol_error)
    {
        result = FAILED;
    }

    return (result);

}                                /* End Of Procedure ir_ASE */

/*****************************************************************************
 *
 * Function: ir_control
 *
 * Purpose: This function polls for an event from the client or from the
 *          target process and then dispatches the event to the appropriate
 *          function ir_ASE() or ACSE() for processing. Firstly, it
 *          creates the client associate information list and then calls
 *          ira_ALI_setup() function.
 *
 * Input Arguments:  serviceTCPPort   - target Z39.50 TCP Port
 *                   tp_hostname      - target process hostname
 *
 * Return values:    SUCCESS          - should never return
 *                   FAILED           - initialization has failed
 *
 * Output Arguments: none
 *
 * Function used:      ira_ALI_setup()
 *                     ira_comm_event()
 *                     ACSE()
 *                     ir_ASE()
 *
 * Version: Date:   93/03/14
 *          Name:   Benjamin Yen
 *          Reason: First Draft of IR Target Kernel
 *
 * Reference: IR KERNEL - TARGET PRELIMINARY SOFTWARE DESIGN DOCUMENT
 *            IR KERNEL - TARGET SOFTWARE DETAIL DESIGN DOCUMENT
 *
 * Notes:
 *
 ****************************************************************************/
int
ir_control(serviceTCPPort, tp_hostname)
    int serviceTCPPort;
    char *tp_hostname;
{

    APDU PDU_ptr;
    int looping_forever, result, pevent, serviceSocket;
    struct queue_struct *assoc_list;
    struct assoc_struct *assocId;

#ifdef DEBUG_LEVEL1
    fprintf(DEBUG_FILE1, "Calling ir_control ...\n");
#endif

    /* Create the multi-threading association control list. */
    if ((assoc_list = ir_create_queue()) == NULL)
    {
        return (FAILED);
    }

    /* Initialize the communication services */
    if (ira_ALI_setup(serviceTCPPort, &serviceSocket) == FAILED)
    {
        return (FAILED);
    }

    looping_forever = TRUE;

    while (looping_forever)
    {
        /* waiting for a communication event to occur */
        do
        {
            pevent = ira_comm_event(serviceSocket, &assoc_list, &assocId, &PDU_ptr);

#ifdef DEBUG_LEVEL3
            if (assocId)
                fprintf(DEBUG_FILE3, "ir_control: assocId->socket_no=%d, event=%d\n", assocId->socket_no[CLIENTSOCK], pevent);
            else
                fprintf(DEBUG_FILE3, "ir_control: No assocId to process\n");
#endif

#ifdef DEBUG_TESTING
            /* slow down the display on screen, it's just for testing purpose */
            if (pevent == NON_EVENT)
            {
                sleep(3);
            }
#endif

        } while (pevent == NON_EVENT);

        while (TRUE)
        {

            /* Dispatch an incoming event to either ir_ASE() or ACSE()  */
            if (confirmed_ACSE(pevent))
            {
                result = ACSE(&assoc_list, &assocId, pevent);
            }
            else if (confirmed_IR(pevent))
            {
                result = ir_ASE(&assoc_list, &assocId, &PDU_ptr, pevent, tp_hostname);
            }

            if (result == FAILED)
            {
                ACSE(&assoc_list, &assocId, A_ABORT_INDICATION);
            }

            if (assocId->APDUsize2 > 0)        /* if 2nd APDU exists */
            {
                pevent = ir_get_second_APDU(&assocId, &PDU_ptr);

#ifdef DEBUG_LEVEL3
                fprintf(DEBUG_FILE3, "event returned from get_second_APDU is %d\n", pevent);
#endif

                if (pevent == 0)
                    break;
            }
            else
                break;                /* no 2nd APDU */
        }
    }                                /* end of while */

    return (SUCCESS);

}                                /* End of Procedure ir_control */


/*****************************************************************************
 *
 * Function: ir_get_parameters
 *
 * Purpose: This function processes the command line parameters to extract
 *          the target service TCP port number and target process hostname.
 *
 * Input Arguments:     argc        - number of arguments.
 *                      argv        - pointer to the command line argument.
 *
 * Return values: None
 *
 * Output Arguments:    target_serviceTCPPort   -  TCP port number for Target
 *                      tp_hostname             -  Target Process hostname.
 *
 * External Variables:
 *
 * Function used:
 *
 * Algorithm:
 *
 * Version: Date:   93/03/14
 *          Name:   Benjamin Yen
 *          Reason: First Draft of IR Target Kernel
 *
 * Reference: IR KERNEL - TARGET PRELIMINARY SOFTWARE DESIGN DOCUMENT
 *            IR KERNEL - TARGET SOFTWARE DETAIL DESIGN DOCUMENT
 *
 * Notes:
 *
 ****************************************************************************/
char *
ir_get_parameters(int argc, char *argv[], int *target_serviceTCPPort)
{

    char *chptr;
    int i;
    char *tp_hostname;

#ifdef DEBUG_LEVEL1
    fprintf(DEBUG_FILE1, "Calling ir_get_parameters ...\n");
#endif

    /* Initialize the output parameter pointers to NULL */
    *target_serviceTCPPort = 0;
    tp_hostname = NULL;

    /*
     * Obtain the target servcie TCP port number and target process hostname
     * from command line arguments e.g.
     * 
     * irtarget host=tempest.acadiau.ca, port=210
     */

    for (i = 1; i < argc; i++)
    {
        chptr = (char *) malloc(strlen(argv[i]) + 1);
        strcpy(chptr, argv[i]);

        /* look for token "port=" */
        if (strcmp(strtok(chptr, "="), "port") == 0)
        {
            *target_serviceTCPPort = atoi(strpbrk(argv[i], "=") + 1);
        }
        else if (strcmp(strtok(chptr, "="), "host") == 0)
        {
            /* look for token  "host=" */
            tp_hostname = (char *) malloc(strlen(strpbrk(argv[i], "=") + 1) + 1);
            strcpy(tp_hostname, strpbrk(argv[i], "=") + 1);
        }

    }                                /* end of for */

    /*
     * if the hostname has not been passed on the command line, set to the
     * default to current hostname
     */
    if (tp_hostname == NULL)
    {
        tp_hostname = (char *) malloc(MAXNAMELENGTH);
        if (gethostname(tp_hostname, MAXNAMELENGTH) < 0)
        {
            return (NULL);
        }
    }

    /*
     * if the service TCP port number has not been passed on the command
     * line, set to the default to the Z39.50 port number (i.e. 210)
     */
    if (*target_serviceTCPPort == 0)
    {
        *target_serviceTCPPort = Z3950_PORT;
    }

    /* Report the command line parameters values on start up */
    printf("Target Service TCP Port = %d\n", *target_serviceTCPPort);
    printf("Target Process Hostname = %s\n", tp_hostname);

    strcpy(TEMP_HOST_NAME, tp_hostname);

    return (tp_hostname);

}                                /* End Of Procedure ir_get_parameters */

/*****************************************************************************
 *
 * Function: ir_open_files
 *
 * Purpose: This function opens the files for debugging and error logging
 *          purposes.
 *
 * Input Arguments:
 *
 * Return values: SUCCESS or FAILED
 *
 * Output Arguments:
 *
 * External Variables:
 *
 * Function used:
 *
 * Algorithm:
 *
 * Version: Date:   93/03/14
 *          Name:   Benjamin Yen
 *          Reason: First Draft of IR Target Kernel
 *
 * Reference: IR KERNEL - TARGET PRELIMINARY SOFTWARE DESIGN DOCUMENT
 *            IR KERNEL - TARGET SOFTWARE DETAIL DESIGN DOCUMENT
 *
 * Notes:
 *
 ****************************************************************************/

ir_open_files()
{
    char tty_no[9];
    FILE *f;

    InitNibbleMem(MAXMESSAGESIZE, MAXMESSAGESIZE);        /* Initialize ASN */

#ifdef DEBUG_LEVEL1

#ifdef VMS
    DEBUG_FILE1 = stdout;
#else
    if ((DEBUG_FILE1 = fopen(DEBUG_FILE_NAME1, "a")) == NULL)
        return (FAILED);
#endif

#endif

#ifdef DEBUG_LEVEL2

#ifdef VMS
    DEBUG_FILE2 = DEBUG_FILE1;
#else
    if ((DEBUG_FILE2 = fopen(DEBUG_FILE_NAME2, "a")) == NULL)
        return (FAILED);
#endif

#endif

#ifdef DEBUG_LEVEL3

#ifdef VMS
    DEBUG_FILE3 = DEBUG_FILE1;
#else
    if ((DEBUG_FILE3 = fopen(DEBUG_FILE_NAME3, "a")) == NULL)
        return (FAILED);
#endif

#endif

    if ((DEBUG_FILE4 = fopen("/dev/null", "a")) == NULL)
        return (FAILED);
	
    return (SUCCESS);

}                                /* End Of Procedure ir_open_files */

/*****************************************************************************
 *
 * Function: ir_open_files
 *
 * Purpose: This function closes the files for debugging and error logging
 *          purposes.
 *
 * Input Arguments:
 *
 * Return values: None
 *
 * Output Arguments:
 *
 * External Variables:
 *
 * Function used:
 *
 * Algorithm:
 *
 * Version: Date:   93/03/14
 *          Name:   Benjamin Yen
 *          Reason: First Draft of IR Target Kernel
 *
 * Reference: IR KERNEL - TARGET PRELIMINARY SOFTWARE DESIGN DOCUMENT
 *            IR KERNEL - TARGET SOFTWARE DETAIL DESIGN DOCUMENT
 *
 * Notes:
 *
 ****************************************************************************/

void
ir_close_files()
{

#ifndef VMS

#ifdef DEBUG_LEVEL1
    fclose(DEBUG_FILE1);
#endif

#ifdef DEBUG_LEVEL2
    fclose(DEBUG_FILE2);
#endif

#ifdef DEBUG_LEVEL3
    fclose(DEBUG_FILE3);
#endif

#endif

}                                /* End of Procedure ir_close_files */

/*****************************************************************************
 *
 * Function: ir_read_default_parameter_file
 *
 * Purpose: This function sets the default parameters and will read the ones
 *            from the default file irtarget.def (if it exists) and reset them.
 *
 * Input Arguments: None
 *
 * Return values: None
 *
 * Output Arguments:
 *
 * External Variables:
 *
 * Function used:
 *
 * Algorithm:
 *
 * Version: Date:   93/03/14
 *          Name:   Benjamin Yen
 *          Reason: First Draft of IR Target Kernel
 *
 * Reference: IR KERNEL - TARGET PRELIMINARY SOFTWARE DESIGN DOCUMENT
 *            IR KERNEL - TARGET SOFTWARE DETAIL DESIGN DOCUMENT
 *
 * Notes:
 *
 ****************************************************************************/
ir_read_default_parameter_file()
{
    FILE *infile;
    int i;
    char str[100];

    strcpy(DEBUG_FILE_NAME1, "irtarget.d1");
    strcpy(DEBUG_FILE_NAME2, "irtarget.d2");
    strcpy(DEBUG_FILE_NAME3, "irtarget.d3");
    strcpy(ERROR_FILE_NAME, "irtarget.err");
    INIT_TIMEOUT_IN_SECONDS = 2 * 60;
    SEARCH_TIMEOUT_IN_SECONDS = 5 * 60;
    PRESENT_TIMEOUT_IN_SECONDS = 5 * 60;
    DELETE_TIMEOUT_IN_SECONDS = 3 * 60;
    TRIGGER_TIMEOUT_IN_SECONDS = 5 * 60;
    ACCESS_TIMEOUT_IN_SECONDS = 2 * 60;
    RESOURCE_TIMEOUT_IN_SECONDS = 3 * 60;
    IDLE_TIMEOUT_IN_SECONDS = 5 * 60;
    MAXMESSAGESIZE = 52000L;
    MAX_ERROR_RECORDS = 1000L;

    if ((infile = fopen("irtarget.def", "r")) != NULL)
    {
        if (fgets(str, 100, infile) == NULL)
        {
            display_error("KERNEL", ERR_INCOMPLETE_DEFAULT_FILE, -1);
            return (FAILED);
        }

        i = 0;
        while (fgets(str, 100, infile) != NULL)
        {
            i++;
            switch (i)
            {
                case 1:
                    sscanf(str, "%s", DEBUG_FILE_NAME1);
                    break;
                case 2:
                    sscanf(str, "%s", DEBUG_FILE_NAME2);
                    break;
                case 3:
                    sscanf(str, "%s", DEBUG_FILE_NAME3);
                    break;
                case 4:
                    sscanf(str, "%s", ERROR_FILE_NAME);
                    break;
                case 5:
                    sscanf(str, "%d", &INIT_TIMEOUT_IN_SECONDS);
                    INIT_TIMEOUT_IN_SECONDS *= 60;
                    break;
                case 6:
                    sscanf(str, "%d", &SEARCH_TIMEOUT_IN_SECONDS);
                    SEARCH_TIMEOUT_IN_SECONDS *= 60;
                    break;
                case 7:
                    sscanf(str, "%d", &PRESENT_TIMEOUT_IN_SECONDS);
                    PRESENT_TIMEOUT_IN_SECONDS *= 60;
                    break;
                case 8:
                    sscanf(str, "%d", &DELETE_TIMEOUT_IN_SECONDS);
                    DELETE_TIMEOUT_IN_SECONDS *= 60;
                    break;
                case 9:
                    sscanf(str, "%d", &TRIGGER_TIMEOUT_IN_SECONDS);
                    TRIGGER_TIMEOUT_IN_SECONDS *= 60;
                    break;
                case 10:
                    sscanf(str, "%d", &ACCESS_TIMEOUT_IN_SECONDS);
                    ACCESS_TIMEOUT_IN_SECONDS *= 60;
                    break;
                case 11:
                    sscanf(str, "%d", &RESOURCE_TIMEOUT_IN_SECONDS);
                    RESOURCE_TIMEOUT_IN_SECONDS *= 60;
                    break;
                case 12:
                    sscanf(str, "%d", &IDLE_TIMEOUT_IN_SECONDS);
                    IDLE_TIMEOUT_IN_SECONDS *= 60;
                    break;
                case 13:
                    sscanf(str, "%ld", &MAXMESSAGESIZE);
                    break;
                case 14:
                    sscanf(str, "%ld", &MAX_ERROR_RECORDS);
                    break;
            }
        }

        if (i != 17)
        {
            printf("%d entries have read from irtarget.def\n", i);
            display_error("KERNEL", ERR_INCOMPLETE_DEFAULT_FILE, -1);
            return (FAILED);
        }

        fclose(infile);
    }

    printf("\n");
    printf("FIRST DEBUG FILE NAME       = %s\n", DEBUG_FILE_NAME1);
    printf("SECOND DEBUG FILE NAME      = %s\n", DEBUG_FILE_NAME2);
    printf("THIRD DEBUG FILE NAME       = %s\n", DEBUG_FILE_NAME3);
    printf("ERROR FILE NAME             = %s\n", ERROR_FILE_NAME);
    printf("INIT TIMEOUT IN MINUTES     = %d\n", INIT_TIMEOUT_IN_SECONDS / 60);
    printf("SEARCH TIMEOUT IN MINUTES   = %d\n", SEARCH_TIMEOUT_IN_SECONDS / 60);
    printf("PRESENT TIMEOUT IN MINUTES  = %d\n", PRESENT_TIMEOUT_IN_SECONDS / 60);
    printf("DELETE TIMEOUT IN MINUTES   = %d\n", DELETE_TIMEOUT_IN_SECONDS / 60);
    printf("TRIGGER TIMEOUT IN MINUTES  = %d\n", TRIGGER_TIMEOUT_IN_SECONDS / 60);
    printf("ACCESS TIMEOUT IN MINUTES   = %d\n", ACCESS_TIMEOUT_IN_SECONDS / 60);
    printf("RESOURCE TIMEOUT IN MINUTES = %d\n", RESOURCE_TIMEOUT_IN_SECONDS / 60);
    printf("IDLE TIMEOUT IN MINUTES     = %d\n", IDLE_TIMEOUT_IN_SECONDS / 60);
    printf("MAX MESSAGE SIZE            = %ld\n", MAXMESSAGESIZE);
    printf("MAX ERROR RECORDS           = %ld\n", MAX_ERROR_RECORDS);
    printf("\n");

    return (SUCCESS);
}

/*****************************************************************************
 *
 * Function: main
 *
 * Purpose: This is the very first function which invokes the ir_control()
 *          function and takes two arguments: TCP port number and the
 *          target process host name.
 *
 * Input Arguments: argc - number of arguments.
 *                  argv - pointer to the command line argument.
 *
 * Return values:   None
 *
 * Output Arguments:
 *
 * External Variables:
 *
 * Function called:    ir_get_parameters()
 *                     ir_open_files()
 *                     ir_control()
 *                     ir_close_files()
 *
 * Algorithm:
 *
 * Version: Date:   93/03/14
 *          Name:   Benjamin Yen
 *          Reason: First Draft of IR Target Kernel
 *
 * Reference: IR KERNEL - TARGET PRELIMINARY SOFTWARE DESIGN DOCUMENT
 *            IR KERNEL - TARGET SOFTWARE DETAIL DESIGN DOCUMENT
 *
 * Notes:
 *
 ****************************************************************************/

#ifdef VMS
int
#else
void
#endif

main(argc, argv)
    int argc;
    char *argv[];
{

    int target_serviceTCPPort;
    char *tp_hostname;

    printf("\nirtarget version %s\n", SOURCE_CODE_VERSION);

    if (ir_read_default_parameter_file() == FAILED)
        exit(1);

    /* Open debugging files. */
    ir_open_files();

    /* Extract the command line parameters. */
    if ((tp_hostname = ir_get_parameters(argc, argv, &target_serviceTCPPort)) == NULL)
    {
        exit(1);
    }

    /* Call the ir_control function to start the IR kernel. */
    if (ir_control(target_serviceTCPPort, tp_hostname) != SUCCESS)
    {
        display_error("KERNEL", ERR_INTERNAL_SYSTEM_ERROR, -1);
    }

    /* Close files. */
    ir_close_files();

    exit(0);

}                                /* End of main */
