/****************************************************************************
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *~                             IR Kernel                                    ~
 *~                         Copyright(C) 1993                                ~
 *~                       Software Kinetics Ltd.                             ~
 *~                        All Rights Reserved                               ~
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Copyright Software Kinetics Ltd.
 * 
 * Software Kinetics Ltd. has granted a licence of this software to
 * National Library of Canada.
 *
 * The National Library of Canada grants to the user a royalty-free
 * non-exclusive licence in perpetuity to use, modify, improve,
 * reproduce and sell the software.
 *
 * The software is provided to the user "as is" without warranty of
 * any kind.  Neither Software Kinetics Ltd. nor the National Library
 * of Canada warrants,guarantees or makes any representation regarding
 * the use, or the results of the use of the software in terms of
 * accuracy,reliability, or otherwise.  The entire risk as results and
 * performance of the software is assumed by the user. Neither
 * Software Kinetics Ltd. nor the National Library of Canada shall be
 * liable for any direct, indirect, consequential or incidental
 * damages arising out the use of software.
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ 
 *
 * Program: tpaccctr.c
 *
 * Purpose: This file contains the access control functions for target process
 *          (Black Box).
 *
 * Invocation:
 *
 * Input Arguments:
 *
 * Output/Exit Values: None
 *
 * Functions used:
 *              tp_fetchPW()
 *              tp_valid_user()
 *              tp_fill_access_request()
 *              tp_create_access_request()
 *              tp_access_control()
 *
 * Version: Date:   93/05/12
 *          Name:   Benjamin Yen
 *          Reason: First Draft of IR Target Kernel
 *
 * Reference: IR KERNEL - TARGET PRELIMINARY SOFTWARE DESIGN DOCUMENT
 *            IR KERNEL - TARGET SOFTWARE DETAIL DESIGN DOCUMENT
 *            IR KERNEL - INTERFACE REQUIREMENTS SPECIFICATION
 *            OSI - ANSI Z39.50 DOCUMENTATION
 *
 * Notes:
 *
 * $Id: tpaccctr.c,v 1.2 1994/02/11 13:37:04 irsuprt Exp irsuprt $
 * $Log: tpaccctr.c,v $
 * Revision 1.2  1994/02/11  13:37:04  irsuprt
 * indef len
 *
 * Revision 1.1  1993/11/01  17:03:51  irsuprt
 * Initial revision
 *
 *
 *****************************************************************************/
#include <stdio.h>
#include <fcntl.h>
#include <pwd.h>
#include <ctype.h>
#include <sys/errno.h>

#include "tp.h"
#include "tpconst.h"
#include "irerror.h"
#include "tpbbox.h"
#include "ir_oid.h"

extern int errno;

#define SHADOW 1
#define GLOBAL_PASSWORD "irtarget"

/*****************************************************************************
 *
 * Function: tp_fetchPW
 *
 * Purpose: This function returns the passwd structure found in the
 *          Unix file /etc/passwd with the user name 'username'.
 *
 * Input Arguments: username - Unix system user name
 *
 * Return values: the passwd structure pointer
 *
 * Output Arguments:
 *
 * External Variables:
 *
 * Function used:
 *
 * Algorithm:
 *
 * Version: Date:   93/05/12
 *          Name:   Benjamin Yen
 *          Reason: First Draft of IR Target Kernel
 *
 * Reference: IR KERNEL - TARGET PRELIMINARY SOFTWARE DESIGN DOCUMENT
 *            IR KERNEL - TARGET SOFTWARE DETAIL DESIGN DOCUMENT
 *            IR KERNEL - INTERFACE REQUIREMENTS SPECIFICATION
 *            OSI - ANSI Z39.50 DOCUMENTATION
 *
 * Notes:
 *
 ****************************************************************************/

struct passwd *
tp_fetchPW(username)
    char *username;
{
    return ((struct passwd *) getpwnam(username));
}                                /* tp_fetchPW */

/*****************************************************************************
 *
 * Function: tp_get_idauth_pass_length
 *
 * Purpose: This function returns the password length from the idAuthentication
 *          of initReq.
 *
 * Input Arguments:
 *          PDU - the Init Request PDU or the Access Control Response PDU
 *
 * Return values: password length
 *
 * Output Arguments:
 *
 * External Variables:
 *
 * Function used:
 *
 * Algorithm:
 *
 * Version: Date:   93/05/12
 *          Name:   Benjamin Yen
 *          Reason: First Draft of IR Target Kernel
 *
 * Reference: IR KERNEL - TARGET PRELIMINARY SOFTWARE DESIGN DOCUMENT
 *            IR KERNEL - TARGET SOFTWARE DETAIL DESIGN DOCUMENT
 *            IR KERNEL - INTERFACE REQUIREMENTS SPECIFICATION
 *            OSI - ANSI Z39.50 DOCUMENTATION
 *
 * Notes:
 *
 ****************************************************************************/
unsigned long int tp_get_idauth_pass_length(PDU)
    APDU *PDU;
{

	if (PDU->a.initReq->idAuthentication.octetLen <= 0)
	{
		#ifdef DEBUG_LEVEL1
			fprintf(DEBUG_FILE1, "BBOX: idAuthentication has no data\n");
		#endif
		return(0);
	}
	else
		return(PDU->a.initReq->idAuthentication.octetLen);

#ifdef NEW_APDU_H
	if (PDU->a.initReq->idAuthentication == NULL)
	{
		#ifdef DEBUG_LEVEL1
			fprintf(DEBUG_FILE1, "BBOX: idAuthentication has no data\n");
		#endif
		return(0);
	}

	switch (PDU->a.initReq->idAuthentication->choiceId)
	{

		case IDAUTHENTICATION_OPEN:
			#ifdef DEBUG_LEVEL1
				fprintf(DEBUG_FILE1, "BBOX: idAuthentication.choiceId=Open\n");
			#endif
			return(PDU->a.initReq->idAuthentication->a.open->octetLen);
			break;

        case IDAUTHENTICATION_USERPASS:
			#ifdef DEBUG_LEVEL1
				fprintf(DEBUG_FILE1, "BBOX: idAuthentication.choiceId=Pass\n");
			#endif
			return(PDU->a.initReq->idAuthentication->a.userPass->password.octetLen);
			break;

        case IDAUTHENTICATION_ANONYMOUS:
			#ifdef DEBUG_LEVEL1
				fprintf(DEBUG_FILE1, "BBOX: idAuthentication.choiceId=Anonymous\n");
			#endif
			return(0); /* just a character so return 0 */
			break;

        case IDAUTHENTICATION_OTHER:
			#ifdef DEBUG_LEVEL1
				fprintf(DEBUG_FILE1, "BBOX: idAuthentication.choiceId=other\n");
			#endif
			return(0); /* not supported */
			break;
			
	};
#endif
}

/*****************************************************************************
 *
 * Function: tp_get_idauth_pass
 *
 * Purpose: This function returns the password from the idAuthentication
 *          of initReq.
 *
 * Input Arguments:
 *          PDU - the Init Request PDU or the Access Control Response PDU
 *
 * Return values: password 
 *
 * Output Arguments:
 *
 * External Variables:
 *
 * Function used:
 *
 * Algorithm:
 *
 * Version: Date:   93/05/12
 *          Name:   Benjamin Yen
 *          Reason: First Draft of IR Target Kernel
 *
 * Reference: IR KERNEL - TARGET PRELIMINARY SOFTWARE DESIGN DOCUMENT
 *            IR KERNEL - TARGET SOFTWARE DETAIL DESIGN DOCUMENT
 *            IR KERNEL - INTERFACE REQUIREMENTS SPECIFICATION
 *            OSI - ANSI Z39.50 DOCUMENTATION
 *
 * Notes:
 *
 ****************************************************************************/
char * tp_get_idauth_pass(PDU)
    APDU *PDU;
{

	if (PDU->a.initReq->idAuthentication.octs == NULL)
	{
		#ifdef DEBUG_LEVEL1
			fprintf(DEBUG_FILE1, "BBOX: idAuthentication has no data\n");
		#endif
		return(NULL);
	}
	else
		return(PDU->a.initReq->idAuthentication.octs);

#ifdef NEW_APDU_H
	if (PDU->a.initReq->idAuthentication == NULL)
	{
		#ifdef DEBUG_LEVEL1
			fprintf(DEBUG_FILE1, "BBOX: idAuthentication has no data\n");
		#endif
		return(NULL);
	}

	switch (PDU->a.initReq->idAuthentication->choiceId)
	{

		case IDAUTHENTICATION_OPEN:
			return(PDU->a.initReq->idAuthentication->a.open->octs);
			break;

        case IDAUTHENTICATION_USERPASS:
			return(PDU->a.initReq->idAuthentication->a.userPass->password.octs);
			break;

        case IDAUTHENTICATION_ANONYMOUS:
			return(NULL); /* just a character so return 0 */
			break;

        case IDAUTHENTICATION_OTHER:
			return(NULL); /* not supported */
			break;
			
	};
#endif
}

/*****************************************************************************
 *
 * Function: tp_valid_user
 *
 * Purpose: This function validates the user password against the global
 *          password; however, the codes surrounded with '#ifdef NOT_USE' can
 *          validate one password per user i.e. every user can have an unique
 *          password (the only restriction is that it only applies to Unix
 *          environment).
 *
 * Input Arguments:
 *          PDU - the Init Request PDU or the Access Control Response PDU
 *          option - tells which PDU's password to use
 *
 * Return values: TRUE - pass the validation of the password
 *                FALSE - fail the validation of the password
 *
 * Output Arguments:
 *
 * External Variables:
 *
 * Function used:
 *
 * Algorithm:
 *
 * Version: Date:   93/05/12
 *          Name:   Benjamin Yen
 *          Reason: First Draft of IR Target Kernel
 *
 * Reference: IR KERNEL - TARGET PRELIMINARY SOFTWARE DESIGN DOCUMENT
 *            IR KERNEL - TARGET SOFTWARE DETAIL DESIGN DOCUMENT
 *            IR KERNEL - INTERFACE REQUIREMENTS SPECIFICATION
 *            OSI - ANSI Z39.50 DOCUMENTATION
 *
 * Notes:
 *
 ****************************************************************************/
tp_valid_user(PDU, option)
    APDU *PDU;
    int option;
{
    char *crypt();
    char *encryptpasswd;
    struct passwd *pwd;
    int i;

    /* must handle shadow passwords */

#ifdef DEBUG_LEVEL1
    fprintf(DEBUG_FILE1, "Calling tp_valid_user\n");
#endif

    /* Checking the password from the INIT request field idauthentication. */
    if (option == CHECK_INIT_PASSWORD)
    {
        if (tp_get_idauth_pass_length(PDU) != strlen(GLOBAL_PASSWORD))
        {
            sprintf(ERR_STR, "PASSWORD LENGTH IS INVALID BUT IT WILL RE-TRY, client=%ld, kernel=%d", tp_get_idauth_pass_length(PDU), strlen(GLOBAL_PASSWORD));
            display_error("BBOX", DISPLAY_USER_ERROR, -1);
            return (FALSE);
        }

        if (memcmp(tp_get_idauth_pass(PDU), GLOBAL_PASSWORD, tp_get_idauth_pass_length(PDU)) == 0)
            return (TRUE);        /* match/right password */
        else
        {
            strcpy(ERR_STR, "PASSWORD IS INVALID");
            display_error("BBOX", DISPLAY_USER_ERROR, -1);
            return (FALSE);
        }
    }

    /* Checking the password from the Access Control Response */
    if (PDU->a.accessControlResp->securityChallengeResponse->choiceId == ACCESSCONTROLRESPCHOICE_OLDWAY)
    {

#ifdef DEBUG_LEVEL1
    fprintf(DEBUG_FILE1, "tp_valid_user: client is using access control OLDWAY\n");
#endif

        /* Using the old data structure. */
        if (PDU->a.accessControlResp->securityChallengeResponse->a.oldway->octetLen != strlen(GLOBAL_PASSWORD))
        {
            sprintf(ERR_STR, "PASSWORD LENGTH IS INVALID, client=kernel, %ld=%d", PDU->a.accessControlResp->securityChallengeResponse->a.oldway->octetLen, strlen(GLOBAL_PASSWORD));
            display_error("BBOX", DISPLAY_USER_ERROR, -1);
            return (FALSE);
        }

        /* Only password, all users uses one password. */
        if (memcmp(PDU->a.accessControlResp->securityChallengeResponse->a.oldway->octs, GLOBAL_PASSWORD, PDU->a.accessControlResp->securityChallengeResponse->a.oldway->octetLen) == 0)
            return (TRUE);        /* match/right password */
        else
        {
            strcpy(ERR_STR, "PASSWORD IS INVALID");
            display_error("BBOX", DISPLAY_USER_ERROR, -1);
            return (FALSE);
        }
    }
    else
    {

#ifdef DEBUG_LEVEL1
    fprintf(DEBUG_FILE1, "tp_valid_user: client is using access control NEWWAY\n");
#endif
        /* Using the new data structure. */
        if (memcmp(OID_CLASS_ACCESS_CONTROL_ENCODED, PDU->a.accessControlResp->securityChallengeResponse->a.newway->direct_reference.octs, PDU->a.accessControlResp->securityChallengeResponse->a.newway->direct_reference.octetLen) != 0)
        {
            strcpy(ERR_STR, "ACCESS CONTROL OBJECT IDENTIFIER IS INVALID");
            display_error("BBOX", DISPLAY_USER_ERROR, -1);
            return (FALSE);
        }

        if (PDU->a.accessControlResp->securityChallengeResponse->a.newway->encoding->choiceId == EXTERNALCHOICE_SINGLE_ASN1_TYPE)
        {
            if (PDU->a.accessControlResp->securityChallengeResponse->a.newway->encoding->a.single_ASN1_type->octetLen != strlen(GLOBAL_PASSWORD))
            {
                sprintf(ERR_STR, "PASSWORD LENGTH IS INVALID, client=kernel, %ld=%d", PDU->a.accessControlResp->securityChallengeResponse->a.newway->encoding->a.single_ASN1_type->octetLen, strlen(GLOBAL_PASSWORD));
                display_error("BBOX", DISPLAY_USER_ERROR, -1);
                return (FALSE);
            }

            if (memcmp(PDU->a.accessControlResp->securityChallengeResponse->a.newway->encoding->a.single_ASN1_type->octs, GLOBAL_PASSWORD, PDU->a.accessControlResp->securityChallengeResponse->a.newway->encoding->a.single_ASN1_type->octetLen) == 0)
                return (TRUE);
            else
            {
                strcpy(ERR_STR, "PASSWORD IS INVALID");
                display_error("BBOX", DISPLAY_USER_ERROR, -1);
                return (FALSE);
            }
        }
        else if (PDU->a.accessControlResp->securityChallengeResponse->a.newway->encoding->choiceId == EXTERNALCHOICE_OCTET_ALIGNED)
        {
            if (PDU->a.accessControlResp->securityChallengeResponse->a.newway->encoding->a.octet_aligned->octetLen != strlen(GLOBAL_PASSWORD))
            {
                sprintf(ERR_STR, "PASSWORD LENGTH IS INVALID, client=kernel, %ld=%d", PDU->a.accessControlResp->securityChallengeResponse->a.newway->encoding->a.octet_aligned->octetLen, strlen(GLOBAL_PASSWORD));
                display_error("BBOX", DISPLAY_USER_ERROR, -1);
                return (FALSE);
            }

            if (memcmp(PDU->a.accessControlResp->securityChallengeResponse->a.newway->encoding->a.octet_aligned->octs, GLOBAL_PASSWORD, PDU->a.accessControlResp->securityChallengeResponse->a.newway->encoding->a.octet_aligned->octetLen) == 0)
                return (TRUE);
            else
            {
                strcpy(ERR_STR, "PASSWORD IS INVALID");
                display_error("BBOX", DISPLAY_USER_ERROR, -1);
                return (FALSE);
            }
        }
        else
            return (FALSE);

#ifdef NOT_USE
        /* User and Password, every user can have an unique password. */
        /* At this time, one global password for all users is being used. */

#ifdef SHADOW
        return (pwdauth(username, password) == 0);
#else
        pwd = tp_fetchPW(username);
        encryptpasswd = crypt(password, pwd->pw_passwd);
        return (strcmp(encryptpasswd, pwd->pw_passwd) == 0);
#endif

#endif
    }

}                                /* tp_valid_user */

/*****************************************************************************
 *
 * Function: tp_fill_access_request
 *
 * Purpose: This function fills the access control request PDU (new data
 *          structure or old data structure depending on the #ifdef
 *          ACCESS_CONTROL_NEWWAY.
 *
 * Input Arguments:
 *
 * Return values: SUCCESS or FAILED
 *
 * Output Arguments:
 *              PDU_ptr - filled and created PDU.
 *
 * External Variables:
 *
 * Function used:
 *              ASN_MakeOid()
 *
 * Algorithm:
 *
 * Version: Date:   93/05/12
 *          Name:   Benjamin Yen
 *          Reason: First Draft of IR Target Kernel
 *
 * Reference: IR KERNEL - TARGET PRELIMINARY SOFTWARE DESIGN DOCUMENT
 *            IR KERNEL - TARGET SOFTWARE DETAIL DESIGN DOCUMENT
 *            IR KERNEL - INTERFACE REQUIREMENTS SPECIFICATION
 *            OSI - ANSI Z39.50 DOCUMENTATION
 *
 * Notes:
 *
 ****************************************************************************/
tp_fill_access_request(PDU_ptr, reference_id_length, reference_id)
    APDU *PDU_ptr;
    AsnInt reference_id_length;
    char *reference_id;
{

    int i;
    AsnOid *temp_oid;

#ifdef DEBUG_LEVEL1
    fprintf(DEBUG_FILE1, "BBOX: calling tp_fill_access_request\n");
#endif

    PDU_ptr->a.accessControlReq = (struct AccessControlReq *) Asn1Alloc(sizeof(struct AccessControlReq));

    PDU_ptr->choiceId = APDU_ACCESSCONTROLREQ;

    PDU_ptr->a.accessControlReq->referenceId.octetLen = reference_id_length;
    PDU_ptr->a.accessControlReq->referenceId.octs = (char *) Asn1Alloc(reference_id_length);
    memcpy(PDU_ptr->a.accessControlReq->referenceId.octs, reference_id, reference_id_length);

    PDU_ptr->a.accessControlReq->securityChallenge = (struct AccessControlReqChoice *) Asn1Alloc(sizeof(struct AccessControlReqChoice));

#ifdef ACCESS_CONTROL_NEWWAY

#ifdef DEBUG_LEVEL1
    fprintf(DEBUG_FILE1, "BBOX: tp_fill_access_request: ACCESS_CONTROL_NEWWAY\n");
#endif

    PDU_ptr->a.accessControlReq->securityChallenge->choiceId = ACCESSCONTROLREQCHOICE_NEWWAY;

    PDU_ptr->a.accessControlReq->securityChallenge->a.newway = (EXTERNAL *) Asn1Alloc(sizeof(EXTERNAL));

    ASN_MakeOid((char *) OID_CLASS_ACCESS_CONTROL, &PDU_ptr->a.accessControlReq->securityChallenge->a.newway->direct_reference);

#ifdef DEBUG_LEVEL2
    fprintf(DEBUG_FILE2, "BBOX: direct_reference=");
    for (i = 0; i < PDU_ptr->a.accessControlReq->securityChallenge->a.newway->direct_reference.octetLen; i++)
        fprintf(DEBUG_FILE2, "%c", PDU_ptr->a.accessControlReq->securityChallenge->a.newway->direct_reference.octs[i]);
    fprintf(DEBUG_FILE2, "\nBBOX: end of direct_reference\n");
#endif

    PDU_ptr->a.accessControlReq->securityChallenge->a.newway->indirect_reference = (AsnInt *) Asn1Alloc(sizeof(AsnInt));
    *PDU_ptr->a.accessControlReq->securityChallenge->a.newway->indirect_reference = 0;

    PDU_ptr->a.accessControlReq->securityChallenge->a.newway->data_value_descriptor.octetLen = 0;
    PDU_ptr->a.accessControlReq->securityChallenge->a.newway->data_value_descriptor.octs = NULL;

    PDU_ptr->a.accessControlReq->securityChallenge->a.newway->encoding = (struct EXTERNALChoice *) Asn1Alloc(sizeof(struct EXTERNALChoice));
    PDU_ptr->a.accessControlReq->securityChallenge->a.newway->encoding->choiceId = EXTERNALCHOICE_OCTET_ALIGNED;
    PDU_ptr->a.accessControlReq->securityChallenge->a.newway->encoding->a.octet_aligned = (AsnOcts *) Asn1Alloc(sizeof(AsnOcts));
    PDU_ptr->a.accessControlReq->securityChallenge->a.newway->encoding->a.octet_aligned->octetLen = sizeof("Please submit a password:");
    PDU_ptr->a.accessControlReq->securityChallenge->a.newway->encoding->a.octet_aligned->octs = (char *) Asn1Alloc(sizeof("Please submit a password:"));
    memcpy(PDU_ptr->a.accessControlReq->securityChallenge->a.newway->encoding->a.octet_aligned->octs, "Please submit a password:", PDU_ptr->a.accessControlReq->securityChallenge->a.newway->encoding->a.octet_aligned->octetLen);

#else

#ifdef DEBUG_LEVEL1
    fprintf(DEBUG_FILE1, "BBOX: tp_fill_access_request: ACCESS_CONTROL_OLDWAY\n");
#endif

    PDU_ptr->a.accessControlReq->securityChallenge->choiceId = ACCESSCONTROLREQCHOICE_OLDWAY;

    PDU_ptr->a.accessControlReq->securityChallenge->a.oldway = (VisibleString *) Asn1Alloc(sizeof(VisibleString));

    PDU_ptr->a.accessControlReq->securityChallenge->a.oldway->octetLen = sizeof("Please submit a password:");
    PDU_ptr->a.accessControlReq->securityChallenge->a.oldway->octs = (char *) Asn1Alloc(sizeof("Please submit a password:"));
    memcpy(PDU_ptr->a.accessControlReq->securityChallenge->a.oldway->octs, "Please submit a password:", PDU_ptr->a.accessControlReq->securityChallenge->a.newway->direct_reference.octetLen);

#endif

    return (SUCCESS);
}

/*****************************************************************************
 *
 * Function: tp_create_access_request
 *
 * Purpose: This function creates the access control request PDU and encodes
 *          it.
 *
 * Input Arguments:
 *          data_ptr - temporary data buffer during encoding process.
 *
 * Return values:
 *
 * Output Arguments:
 *          outputBuf - encoded APDU from the created PDU.
 *          APDU_length - the length of the encoded APDU.
 *
 * External Variables:
 *
 * Function used:
 *          tp_fill_access_request()
 *          SBufInstallData()
 *          SBufResetInWriteRvsMode()
 *          BEncAPDU()
 *          ir_fwrite()
 *
 * Algorithm:
 *
 * Version: Date:   93/05/12
 *          Name:   Benjamin Yen
 *          Reason: First Draft of IR Target Kernel
 *
 * Reference: IR KERNEL - TARGET PRELIMINARY SOFTWARE DESIGN DOCUMENT
 *            IR KERNEL - TARGET SOFTWARE DETAIL DESIGN DOCUMENT
 *            IR KERNEL - INTERFACE REQUIREMENTS SPECIFICATION
 *            OSI - ANSI Z39.50 DOCUMENTATION
 *
 * Notes:
 *
 ****************************************************************************/
tp_create_access_request(data_ptr, outputBuf, APDU_length, reference_id_length, reference_id)
    char **data_ptr;
    SBuf *outputBuf;
    int *APDU_length;
    AsnInt reference_id_length;
    char *reference_id;
{
    APDU PDU_ptr;
    int i;

#ifdef DEBUG_LEVEL1
    fprintf(DEBUG_FILE1, "BBOX:Calling tp_create_access_request()\n");
#endif

    tp_fill_access_request(&PDU_ptr, reference_id_length, reference_id);

    SBufInstallData(outputBuf, (*data_ptr), MAXMESSAGESIZE * 2);
    SBufResetInWriteRvsMode(outputBuf);

    *APDU_length = BEncAPDU(outputBuf, &PDU_ptr);

	if (SBufWriteError(outputBuf))
	{
		display_error("KERNEL", ERR_ENCODE_APDU, -1);
		return(FAILED);
	}

#ifdef DEBUG_LEVEL2
    fprintf(DEBUG_FILE2, "BBOX:Access Request APDU Dump from tpaccctr\n");
    ir_fwrite(SBufDataPtr(outputBuf), SBufDataLen(outputBuf), 1, DEBUG_FILE2);
#endif

    return (SUCCESS);
}

/*****************************************************************************
 *
 * Function: tp_access_control
 *
 * Purpose: This function creates an access control request PDU, encodes it
 *          and then send to IR kernel. Secondly, it tries to receive the
 *          access control response from IR kernel.
 *
 * Input Arguments:
 *              socketFd - socket Fd to send/receive the access control APDU
 *                         to/from.
 *              data_ptr - temporary data buffer used by the ASN functions.
 *
 * Return values: SUCCESS or FAILED
 *
 * Output Arguments:
 *
 * External Variables:
 *
 * Function used:
 *              tp_create_access_request()
 *              tpa_send_APDU()
 *              display_error()
 *              tpa_service_event()
 *              tp_valid_user()
 *
 * Algorithm:
 *
 * Version: Date:   93/05/12
 *          Name:   Benjamin Yen
 *          Reason: First Draft of IR Target Kernel
 *
 * Reference: IR KERNEL - TARGET PRELIMINARY SOFTWARE DESIGN DOCUMENT
 *            IR KERNEL - TARGET SOFTWARE DETAIL DESIGN DOCUMENT
 *            IR KERNEL - INTERFACE REQUIREMENTS SPECIFICATION
 *            OSI - ANSI Z39.50 DOCUMENTATION
 *
 * Notes:
 *
 ****************************************************************************/
tp_access_control(socketFd, data_ptr, reference_id_length, reference_id)
    int socketFd;
    char **data_ptr;
    AsnInt reference_id_length;
    char *reference_id;
{
    int result;
    long APDU_length;
    SBuf APDU_ptr;
    APDU PDU;

#ifdef DEBUG_LEVEL1
    fprintf(DEBUG_FILE1, "BBOX:Calling tp_access_control()\n");
#endif

    if (!ACCESS_CONTROL_REQUIRED)
        return (SUCCESS);

    /* create an access control request APDU */
    tp_create_access_request(data_ptr, &APDU_ptr, &APDU_length, reference_id_length, reference_id);

    /* send an access control request to IR kernel */
    if (tpa_send_APDU(socketFd, &APDU_ptr, APDU_length) <= 0)
    {
        display_error("TP_BBOX", ERR_SEND_DATA_TO_IR_FROM_TP, errno);
        return (FAILED);
    }

    /* receive the access control response from IR Kernel */
    if (tpa_service_event(socketFd, &PDU) == SUCCESS)
    {
        /* validate the password */
        if (tp_valid_user(&PDU, CHECK_ACCESS_PASSWORD))
            return (SUCCESS);
        else
            return (FAILED);
    }

    return (SUCCESS);
}
